import re, time
import numpy as np
import warnings

import zhinst.ziPython as ziPython
import zhinst.utils as utils

def run_example(port=None, api_level=None, do_plot=False):
    """
    Obtain boxcar data using ziDAQServer's blocking (synchronous) poll()
    command.
    """
    # open a connection to a Zurich Instruments server
    daq = utils.autoConnect(port, api_level)
    # detect a device
    device = utils.autoDetect(daq)
    # find out whether the device is an HF2 or a UHF
    devtype = daq.getByte('/%s/features/devtype' %device)
    options = daq.getByte('/%s/features/options' %device)

    if not re.search('BOX',options):
        raise Exception("This example can only be ran on a UHF with the BOX option enabled.")

    if daq.getConnectionAPILevel() != 4:
        warnings.warn("ziDAQServer is using API Level 1, it is strongly recommended to use API Level 4 in order to obtain boxcar data with timestamps.")

    # define configuration and configure the device
    in_c = 0
    out_c = 0
    osc_c = 0
    boxcar_c = 0
    inputpwa_c = 0
    amplitude = 0.5
    frequency = 9.11e6
    windowstart = 75  # boxcar windowstart
    windowsize = 10   # boxcar windowsize
    if re.match('UHF',devtype) and (not re.search('MF',options)):
        out_mixer_c = 3
    else:
        out_mixer_c = 0

    # create a base configuration: disable all outputs, demods and scopes
    general_setting = [
         ['/%s/demods/*/rate' %device, 0],
         ['/%s/demods/*/trigger' %device, 0],
         ['/%s/sigouts/*/enables/*' %device, 0],
         ['/%s/demods/*/enable' %device, 0],
         ['/%s/scopes/*/enable' %device, 0] ]
    daq.set(general_setting)

    # now the settings relevant to this experiment
    exp_setting = [ 
        ['/%s/sigins/%d/imp50' %(device,in_c), 1],
        ['/%s/sigins/%d/ac'    %(device,in_c), 0],
        ['/%s/sigins/%d/range' %(device,in_c), 2*amplitude],
        
        ['/%s/inputpwas/%d/oscselect'   %(device, inputpwa_c), osc_c],
        ['/%s/inputpwas/%d/inputselect' %(device, inputpwa_c), in_c],
        ['/%s/inputpwas/%d/mode'        %(device, inputpwa_c), 1],
        ['/%s/inputpwas/%d/shift'       %(device, inputpwa_c), 0.0],
        ['/%s/inputpwas/%d/harmonic'    %(device, inputpwa_c), 1],
        ['/%s/inputpwas/%d/enable'      %(device, inputpwa_c), 1],

        ['/%s/boxcars/%d/oscselect'   %(device, boxcar_c), osc_c],
        ['/%s/boxcars/%d/adcselect'   %(device, boxcar_c), in_c],
        ['/%s/boxcars/%d/windowstart' %(device, boxcar_c), windowstart],
        ['/%s/boxcars/%d/windowsize'  %(device, boxcar_c), windowsize],
        ['/%s/boxcars/%d/targetrate'  %(device, boxcar_c), 1e3],
        ['/%s/boxcars/%d/periods'     %(device, boxcar_c), 1],
        ['/%s/boxcars/%d/enable'      %(device, boxcar_c), 1],

        ['/%s/oscs/%d/freq'        %(device,osc_c), frequency],

        ['/%s/sigouts/%d/on'            %(device,out_c), 1],
        ['/%s/sigouts/%d/enables/%d'    %(device,out_c,out_mixer_c), 1],
        ['/%s/sigouts/%d/range'         %(device,out_c), 1],
        ['/%s/sigouts/%d/amplitudes/%d' %(device,out_c,out_mixer_c), amplitude]
    ]
    daq.set(exp_setting);

    # unsubscribe any data
    daq.unsubscribe('*'); time.sleep(1.0)

    # clear data buffers, sleep to ensure they are cleared before (re)subscribing
    daq.flush(); time.sleep(1.0)


    paths = ['/%s/boxcars/%d/sample'  % (device, boxcar_c),
             '/%s/boxcars/%d/periods' % (device, boxcar_c),
             '/%s/inputpwas/%d/wave'  % (device, inputpwa_c)]
    daq.subscribe(paths)
    # getAsEvent(): make sure we get the first value even if it hasn't changed.
    daq.getAsEvent('/%s/boxcars/%d/periods' %(device, boxcar_c))

    for periods in np.logspace(0,9,10,base=2):
        time.sleep(1.0)
        daq.setInt('/%s/boxcars/%d/periods' %(device, boxcar_c), int(periods))

    # poll data
    poll_length = 0.1 # [s]
    poll_timeout = 500 # [ms]
    data = daq.poll( poll_length, poll_timeout );
    # unsubscribe from all paths
    daq.unsubscribe('*')
  
    # check the dictionary returned is non-empty
    assert data, "poll returned an empty data dictionary, did you subscribe to any paths?"
    assert device in data, "data dictionary has no key '%s'" %(device)
    assert 'boxcars' in data[device], "data[device] dictionary has no key 'boxcars'"
    assert 'inputpwas' in data[device], "data[device] dictionary has no key 'inputpwas'"

    # the data returned is a dictionary of dictionaries that reflects the node's path
    boxcar_sample = data[device]['boxcars'][str(boxcar_c)]['sample']

    if daq.getConnectionAPILevel() == 4:
        # When using API Level 4 poll() returns both the 'value' and
        # 'timestamp' of the node. These are two vectors of the same length;
        # which consist of (timestamp,value) pairs.
        boxcar_value     = data[device]['boxcars'][str(boxcar_c)]['sample']['value']
        boxcar_timestamp = data[device]['boxcars'][str(boxcar_c)]['sample']['timestamp']
        boxcar_periods_value     = data[device]['boxcars'][str(boxcar_c)]['periods']['value']
        boxcar_periods_timestamp = data[device]['boxcars'][str(boxcar_c)]['periods']['timestamp']
    elif daq.getConnectionAPILevel() == 1:
        # When using API Level 1 poll() returns only the 'value' in the node itself.
        boxcar_value = data[device]['boxcars'][str(boxcar_c)]['sample']
    else:
        raise Exception("Invalid API Level, this should never happen")

    print "Measured average boxcar amplitude is %.5eVs." % np.mean(boxcar_value)

    if do_plot:
        import matplotlib
        import matplotlib.pyplot as plt

        if daq.getConnectionAPILevel() == 4:
            # get the sample rate of the device's ADCs
            clockbase = float(daq.getInt('/%s/clockbase' % device))
            # convert timestamps from ticks to seconds via clockbase
            boxcar_t = (boxcar_timestamp - boxcar_timestamp[0])/clockbase
            boxcar_periods_t = (boxcar_periods_timestamp - boxcar_periods_timestamp[0])/clockbase
            boxcar_periods_t[0] = boxcar_t[0]

        # Create plot 
        fig = plt.figure()
        ax = fig.add_subplot(111)
        plt.grid(True)
        if daq.getConnectionAPILevel() == 4:
            ax.plot(boxcar_t, boxcar_value, label = 'boxcar output')
            plt.xlabel('Time (s)')
            ax2 = ax.twinx()
            ax2.step(np.append(boxcar_periods_t, boxcar_t[-1]), np.append(boxcar_periods_value, boxcar_periods_value[-1]), '-r', label = 'Averaging periods')
            ax2.set_yscale("log")
            plt.xlim(min(boxcar_t[0], boxcar_periods_t[0]), max(boxcar_t[-1], boxcar_periods_t[-1]))
            ax2.legend(loc=1)
        else:
            plt.plot(boxcar_value)
            plt.xlabel('Samples: Use API Level 4 for timestamps!')
            plt.xlim(0, len(boxcar_value))
        plt.title('Boxcar output: The effect of averaging\nperiods on the boxcar value.')
        ax.legend(loc=4)
        ax.set_ylabel('Boxcar value (Vs)')
        ax2.set_ylabel('Number of Averaging Periods')

        plt.figure()
        plt.grid(True)
        pwa_wave =  data[device]['inputpwas'][str(inputpwa_c)]['wave'][-1]
        
        pwa_wave['binphase'] = pwa_wave['binphase']*360/(2*np.pi)
        plt.axhline(0, color='k')
        # The inputpwa waveform is stored in 'x', currently 'y' is unused.
        plt.plot(pwa_wave['binphase'],pwa_wave['x'])
        phase_window = (pwa_wave['binphase'] > windowstart) & (pwa_wave['binphase'] < windowstart+windowsize)
        plt.fill_between(pwa_wave['binphase'], 0, pwa_wave['x'], where=phase_window, alpha=0.5)
        plt.xlim(0, 360)
        plt.title('Input PWA waveform, the shaded region shows the portion\n of the waveform that the boxcar is integrating.')
        plt.xlabel('Phase (degrees)')
        plt.ylabel('Amplitude (V)')

        plt.show()

    return boxcar_sample

if __name__ == '__main__':
    sample = run_example(8004,4,do_plot=True)
